
//
// This source file is part of appleseed.
// Visit http://appleseedhq.net/ for additional information and resources.
//
// This software is released under the MIT license.
//
// Copyright (c) 2010-2013 Francois Beaune, Jupiter Jazz Limited
// Copyright (c) 2014 Francois Beaune, The appleseedhq Organization
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
//

// Interface header.
#include "entitybrowser.h"

// appleseed.renderer headers.
#include "renderer/api/bsdf.h"
#include "renderer/api/color.h"
#include "renderer/api/edf.h"
#include "renderer/api/environmentedf.h"
#include "renderer/api/environmentshader.h"
#include "renderer/api/material.h"
#include "renderer/api/scene.h"
#include "renderer/api/surfaceshader.h"

// appleseed.foundation headers.
#include "foundation/utility/containers/dictionary.h"
#include "foundation/utility/foreach.h"

// Standard headers.
#include <cassert>

using namespace foundation;
using namespace renderer;
using namespace std;

namespace appleseed {
namespace studio {

namespace
{
    template <typename EntityContainer>
    StringDictionary build_entity_dictionary(const EntityContainer& entities)
    {
        StringDictionary result;

        for (const_each<EntityContainer> i = entities; i; ++i)
            result.insert(i->get_name(), i->get_name());

        return result;
    }

    void merge_dictionary(StringDictionary& dest, const StringDictionary& other)
    {
        for (const_each<StringDictionary> i = other; i; ++i)
            dest.insert(i->name(), i->value());
    }

    const Scene* get_parent_scene(const Entity* entity)
    {
        assert(entity);

        while (dynamic_cast<const Scene*>(entity) == 0)
            entity = entity->get_parent();

        return static_cast<const Scene*>(entity);
    }
}


//
// EntityBrowser<Project> class implementation.
//

EntityBrowser<Project>::EntityBrowser(const Project& project)
  : m_project(project)
{
}

StringDictionary EntityBrowser<Project>::get_entities(const string& type) const
{
    return StringDictionary();
}


//
// EntityBrowser<BaseGroup> class implementation.
//

EntityBrowser<BaseGroup>::EntityBrowser(const BaseGroup& base_group)
  : m_base_group(base_group)
{
}

StringDictionary EntityBrowser<BaseGroup>::get_entities(const string& type) const
{
    if (type == "color")
    {
        return build_entity_dictionary(m_base_group.colors());
    }
    else if (type == "texture_instance")
    {
        return build_entity_dictionary(m_base_group.texture_instances());
    }
    else
    {
        return StringDictionary();
    }
}


//
// EntityBrowser<Scene> class implementation.
//

EntityBrowser<Scene>::EntityBrowser(const Scene& scene)
  : EntityBrowser<BaseGroup>(scene)
  , m_scene(scene)
{
}

StringDictionary EntityBrowser<Scene>::get_entities(const string& type) const
{
    if (type == "environment_edf")
    {
        return build_entity_dictionary(m_scene.environment_edfs());
    }
    else if (type == "environment_shader")
    {
        return build_entity_dictionary(m_scene.environment_shaders());
    }
    else
    {
        return EntityBrowser<BaseGroup>::get_entities(type);
    }
}


//
// EntityBrowser<Assembly> class implementation.
//

EntityBrowser<Assembly>::EntityBrowser(const Assembly& assembly)
  : EntityBrowser<BaseGroup>(assembly)
  , m_assembly(assembly)
{
}

StringDictionary EntityBrowser<Assembly>::get_entities(const string& type) const
{
    return get_entities(m_assembly, type);
}

StringDictionary EntityBrowser<Assembly>::get_entities(const Assembly& assembly, const string& type) const
{
    StringDictionary entities;

    if (type == "bsdf")
    {
        entities = build_entity_dictionary(assembly.bsdfs());
    }
    else if (type == "edf")
    {
        entities = build_entity_dictionary(assembly.edfs());
    }
    else if (type == "material")
    {
        entities = build_entity_dictionary(assembly.materials());
    }
    else if (type == "surface_shader")
    {
        entities = build_entity_dictionary(assembly.surface_shaders());
    }
    else if (type == "environment_edf")
    {
        entities = build_entity_dictionary(get_parent_scene(&assembly)->environment_edfs());
    }
    else if (type == "environment_shader")
    {
        entities = build_entity_dictionary(get_parent_scene(&assembly)->environment_shaders());
    }
    else
    {
        entities = EntityBrowser<BaseGroup>::get_entities(type);
    }

    const Assembly* parent_assembly = dynamic_cast<const Assembly*>(assembly.get_parent());

    if (parent_assembly)
    {
        merge_dictionary(entities, get_entities(*parent_assembly, type));
    }
    else
    {
        const Scene* parent_scene = dynamic_cast<const Scene*>(assembly.get_parent());
        assert(parent_scene);

        merge_dictionary(entities, EntityBrowser<Scene>(*parent_scene).get_entities(type));
    }

    return entities;
}

}   // namespace studio
}   // namespace appleseed
